// C/C++ Headers

#include <iostream.h>
#include <string.h>

// Local Headers

#include "ComplexVariables.h"
#include "tf.h"

/*****************************************************************************
 * Initialize static members of class 'ComplexVariables'
 *****************************************************************************/

const dcomplex ComplexVariables::ZERO(0.0, 0.0);

/*****************************************************************************
 * Declaration of private struct 'ComplexEntry'
 *****************************************************************************/

struct ComplexEntry
{
   ComplexEntry(const char *name);
   ~ComplexEntry();

   const char *mName;
   dcomplex mValue;

   ostream &print(ostream &out) const;
};

ostream &operator<<(ostream &out, const ComplexEntry &entry);

/*****************************************************************************
 * Definition of private struct 'ComplexEntry'
 *****************************************************************************/

ComplexEntry::ComplexEntry(const char *name) :
   mName(strcpy(new char[strlen(name) + 1], name)),
   mValue(ComplexVariables::ZERO)
{
}

ComplexEntry::~ComplexEntry()
{
   delete [] (char *)mName;
}

ostream &ComplexEntry::print(ostream &out) const
{
   out << mName << " = [" << __REAL(mValue) << ", " << __IMAG(mValue) << "]";
   return(out);
}

ostream &operator<<(ostream &out, const ComplexEntry &entry)
{
   return(entry.print(out));
}

/*****************************************************************************
 * Definition of class 'ComplexVariables'
 *****************************************************************************/

ComplexVariables::ComplexVariables() :
   mCount(0),
   mEntries(0)
{
}

ComplexVariables::~ComplexVariables()
{
   for(int i = 0; i < mCount; i ++)
      delete mEntries[i];

   delete [] mEntries;
}

int ComplexVariables::assign(const char *name)
{
   int handle = lookup(name);

   if(handle == UNDEFINED)
   {
      ComplexEntry **newEntries = new ComplexEntry *[mCount + 1];

      for(int i = 0; i < mCount; i ++)
         newEntries[i] = mEntries[i];

      newEntries[mCount] = new ComplexEntry(name);
      delete [] mEntries;
      mEntries           = newEntries;

      handle = mCount++;
   }

   return(handle);
}

int ComplexVariables::lookup(const char *name) const
{
   for(int i = 0; i < mCount; i ++)
      if(strcmp(name, mEntries[i]->mName) == 0)
         return(i);

   return(UNDEFINED);
}

const char *ComplexVariables::getName(int handle) const
{
   if(!checkHandle(handle))
      return("null");

   return(mEntries[handle]->mName);
}

void ComplexVariables::initialize()
{
   for(int i = 0; i < mCount; i ++)
      set(i, ZERO);
}

void ComplexVariables::set(int handle, dcomplex value)
{
   if(checkHandle(handle))
      mEntries[handle]->mValue = value;
}

dcomplex ComplexVariables::get(int handle) const
{
   return(checkHandle(handle) ? mEntries[handle]->mValue : ZERO);
}

ostream &ComplexVariables::print(ostream &out) const
{
   for(int i = 0; i < mCount; i ++)
   {
      if(i > 0)
         out << "\n";

      out << "entry[" << i << "]" << " => ";
      mEntries[i]->print(out);
   }

   return(out);
}

int ComplexVariables::main(int, char **)
{
   ComplexVariables variables;

   int xHandle = variables.assign("x");
   cout << "xHandle is: [" << xHandle << "]" << endl;

   int yHandle = variables.assign("y");
   cout << "yHandle is: [" << yHandle << "]" << endl;

   int zHandle = variables.assign("z");
   cout << "zHandle is: [" << zHandle << "]" << endl;

   cout << "name for xHandle is: [" <<
      variables.getName(xHandle) << "]" << endl;

   cout << "name for yHandle is: [" <<
      variables.getName(yHandle) << "]" << endl;

   cout << "name for zHandle is: [" <<
      variables.getName(zHandle) << "]" << endl;

   cout << "lookup(z) => [" << variables.lookup("z") << "]" << endl;

   cout << "lookup(a) => [" << variables.lookup("a") << "]" << endl;

   variables.set(xHandle, dcomplex(1.0, 2.0));
   variables.set(yHandle, dcomplex(3.0, 4.0));
   variables.set(zHandle, dcomplex(5.0, 6.0));

   cout << variables << endl;

   variables.initialize();

   cout << variables << endl;

   return(0);
}

int ComplexVariables::checkHandle(int handle) const
{
   if(handle >= 0 && handle < mCount)
      return(TRUE);

   cout <<
      "ComplexVariables::checkHandle(" << handle << ") - invalid" << endl;

   return(FALSE);
}

ostream &operator<<(ostream &out, const ComplexVariables &cv)
{
   return(cv.print(out));
}

#ifdef __MAIN__

int main(int argc, char **argv)
{
   return(ComplexVariables::main(argc, argv));

}  /* main */

#endif
